/*
  boost_accumulators_distributions.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <boost/random.hpp>
#include <boost/accumulators/accumulators.hpp>
#include <boost/accumulators/statistics/stats.hpp>
#include <boost/accumulators/statistics/mean.hpp>
#include <boost/accumulators/statistics/moment.hpp>
#include <boost/accumulators/statistics/variance.hpp>
#include <boost/accumulators/statistics/skewness.hpp>
#include <boost/accumulators/statistics/kurtosis.hpp>
#include <boost/accumulators/statistics/min.hpp>
#include <boost/accumulators/statistics/max.hpp>
#include <boost/accumulators/statistics/error_of.hpp>
#include <boost/accumulators/statistics/error_of_mean.hpp>
#include <boost/accumulators/statistics/median.hpp>
#include <boost/accumulators/statistics/extended_p_square.hpp>
#include <boost/accumulators/statistics/extended_p_square_quantile.hpp>
#include "statistics.hpp"
#include "statistics_online.hpp"
#include "quantiles.hpp"
#include "quantiles_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

template <typename InputIterator>
std::vector<std::string> tabs_for_string_list(const InputIterator first, const InputIterator last, const ptrdiff_t offset = 0)
{
  std::vector<std::string> rv;
  InputIterator it(first);
  size_t l = std::string(*it++).size();
  while (it != last)
    l = std::max(l, std::string(*it++).size());
  for (it = first; it != last; ++it) {
    rv.push_back(std::string("\t"));
    for (size_t j=0; j<(l+offset)/8 - (std::string(*it).size()+offset)/8; ++j)
      rv[rv.size()-1] += "\t";
  }
  return rv;
}

class distributions {
public:
  typedef boost::mt11213b random_generator_type;
  distributions() :
    random_generator(random_generator_type()),
    uniform_distribution(boost::uniform_01<>()),
    uniform_generator(boost::variate_generator<random_generator_type &,		boost::uniform_01<> >(random_generator, uniform_distribution)),
    gaussian_distribution(boost::normal_distribution<>()),
    gaussian_generator(boost::variate_generator<random_generator_type &,	boost::normal_distribution<> >(random_generator, gaussian_distribution)),
    exponential_distribution(boost::exponential_distribution<>(1)),
    exponential_generator(boost::variate_generator<random_generator_type &,	boost::exponential_distribution<> >(random_generator, exponential_distribution)),
    cauchy_distribution(boost::cauchy_distribution<>(0, 1)),
    cauchy_generator(boost::variate_generator<random_generator_type &,		boost::cauchy_distribution<> >(random_generator, cauchy_distribution)),
    gamma_distribution(boost::gamma_distribution<>(1)),
    gamma_generator(boost::variate_generator<random_generator_type &,		boost::gamma_distribution<> >(random_generator, gamma_distribution)),
    lognormal_distribution(boost::lognormal_distribution<>(1, 1)),
    lognormal_generator(boost::variate_generator<random_generator_type &,	boost::lognormal_distribution<> >(random_generator, lognormal_distribution)),
    poisson_distribution(boost::poisson_distribution<>(2.5)),
    poisson_generator(boost::variate_generator<random_generator_type &,		boost::poisson_distribution<> >(random_generator, poisson_distribution)),
    bernoulli_distribution(boost::bernoulli_distribution<>(.7)),
    bernoulli_generator(boost::variate_generator<random_generator_type &,	boost::bernoulli_distribution<> >(random_generator, bernoulli_distribution)),
    binomial_distribution(boost::binomial_distribution<>(9, .5)),
    binomial_generator(boost::variate_generator<random_generator_type &,	boost::binomial_distribution<> >(random_generator, binomial_distribution)),
    geometric_distribution(boost::geometric_distribution<>(.5)),
    geometric_generator(boost::variate_generator<random_generator_type &,	boost::geometric_distribution<> >(random_generator, geometric_distribution))
  {}
  void seed(uint32_t seed) { random_generator.seed(seed); }
  double generate(int type)
  {
    switch (type) {
    case 0:
      return uniform_generator();
    case 1:
      return gaussian_generator();
    case 2:
      return exponential_generator();
    case 3:
      return cauchy_generator();
    case 4:
      return gamma_generator();
    case 5:
      return lognormal_generator();
    case 6:
      return poisson_generator();
    case 7:
      return bernoulli_generator();
    case 8:
      return binomial_generator();
    default:
      return geometric_generator();
    }      
  }
private:
  random_generator_type random_generator;
  boost::uniform_01<> uniform_distribution;
  boost::variate_generator<random_generator_type &, boost::uniform_01<> > uniform_generator;
  boost::normal_distribution<> gaussian_distribution;
  boost::variate_generator<random_generator_type &, boost::normal_distribution<> > gaussian_generator;
  boost::exponential_distribution<> exponential_distribution;
  boost::variate_generator<random_generator_type &, boost::exponential_distribution<> > exponential_generator;
  boost::cauchy_distribution<> cauchy_distribution;
  boost::variate_generator<random_generator_type &, boost::cauchy_distribution<> > cauchy_generator;
  boost::gamma_distribution<> gamma_distribution;
  boost::variate_generator<random_generator_type &, boost::gamma_distribution<> > gamma_generator;
  boost::lognormal_distribution<> lognormal_distribution;
  boost::variate_generator<random_generator_type &, boost::lognormal_distribution<> > lognormal_generator;
  boost::poisson_distribution<> poisson_distribution;
  boost::variate_generator<random_generator_type &, boost::poisson_distribution<> > poisson_generator;
  boost::bernoulli_distribution<> bernoulli_distribution;
  boost::variate_generator<random_generator_type &, boost::bernoulli_distribution<> > bernoulli_generator;
  boost::binomial_distribution<> binomial_distribution;
  boost::variate_generator<random_generator_type &, boost::binomial_distribution<> > binomial_generator;
  boost::geometric_distribution<> geometric_distribution;
  boost::variate_generator<random_generator_type &, boost::geometric_distribution<> > geometric_generator;
};

const char *distribution_names[] = {
  "uniform",
  "gaussian",
  "exponential",
  "cauchy",
  "gamma",
  "lognormal",
  "poisson",
  "bernoulli",
  "binomial",
  "geometric",
};

int main()
{
  const double boost_maximum_permissible_error(
					       //std::numeric_limits<double>::min()
					       //1e-15 // second raw moment
					       //1e-14 // variance
					       //1e-4 // skewness
					       1e-1 // kurtosis
					       );
  distributions distributions;
  
  const char *result_names[] = {
    "mean",
    "second raw moment",
    "variance",
    "lazy variance",
    "unbiased variance",
    "lazy unbiased variance",
    "third raw moment",
    "skewness",
    "fourth raw moment",
    "kurtosis",
    "min",
    "max",
    "error of mean",
    "median(p^2 quantile)",
    "median(density)",
    "median(p^2 cumulative distribution)",
    "p=0.25 quantile",
    "p=0.5 quantile",
    "p=0.75 quantile",
    "p=0.25 (extended p^2) quantile",
    "p=0.5 (extended p^2) quantile",
    "p=0.75 (extended p^2) quantile",
    "p=0.25 (extended p^2, linear) quantile",
    "p=0.5 (extended p^2, linear) quantile",
    "p=0.75 (extended p^2, linear) quantile",
    "p=0.25 (extended p^2, quadratic) quantile",
    "p=0.5 (extended p^2, quadratic) quantile",
    "p=0.75 (extended p^2, quadratic) quantile",
  };

  std::vector<std::string> tabs_for_result_names(tabs_for_string_list(result_names, result_names + numberof(result_names), 1));
  
  for (int t=0; t<10; ++t) {
    std::cout << distribution_names[t] << ":" << std::endl;
    double probabilities[] = { .25, .5, .75, };
    double detail_probabilities[] = { .1, .2, .3, .4, .5, .6, .7, .8, .9, };
    boost::accumulators::accumulator_set<
      double,
      boost::accumulators::stats<
	boost::accumulators::tag::mean,
	boost::accumulators::tag::moment<2>,
	boost::accumulators::tag::variance(boost::accumulators::immediate),
	boost::accumulators::tag::variance(boost::accumulators::lazy),
	boost::accumulators::tag::moment<3>,
	boost::accumulators::tag::skewness,
	boost::accumulators::tag::moment<4>,
	boost::accumulators::tag::kurtosis,
	boost::accumulators::tag::min,
	boost::accumulators::tag::max,
	boost::accumulators::tag::error_of<boost::accumulators::tag::mean>,
	boost::accumulators::tag::median(boost::accumulators::with_p_square_quantile),
	boost::accumulators::tag::median(boost::accumulators::with_density),
	boost::accumulators::tag::median(boost::accumulators::with_p_square_cumulative_distribution),
	boost::accumulators::tag::extended_p_square//,
	>
      > accumulator_set(
			boost::accumulators::density_cache_size = 10000, boost::accumulators::density_num_bins = 1000,
			boost::accumulators::p_square_cumulative_distribution_num_cells = 100,
			boost::accumulators::tag::extended_p_square::probabilities = probabilities//,
			);
    boost::accumulators::accumulator_set<
      double,
      boost::accumulators::stats<
	boost::accumulators::tag::p_square_quantile
	>
      >
      accumulator_p_quantile_1(boost::accumulators::quantile_probability = 0.25),
      accumulator_p_quantile_2(boost::accumulators::quantile_probability = 0.5),
      accumulator_p_quantile_3(boost::accumulators::quantile_probability = 0.75);
    boost::accumulators::accumulator_set<
      double,
      boost::accumulators::stats<
	boost::accumulators::tag::extended_p_square_quantile
	>
      > accumulator_p_quantiles_linear(
				       boost::accumulators::extended_p_square_probabilities = detail_probabilities
				       );
    boost::accumulators::accumulator_set<
      double,
      boost::accumulators::stats<
	boost::accumulators::tag::extended_p_square_quantile(boost::accumulators::quadratic)
	>
      > accumulator_p_quantiles_quadratic(
					  boost::accumulators::extended_p_square_probabilities = detail_probabilities
					  );
    heuristic_quantiles_online<double> quantiles_online(probabilities), interpolate_quantiles_online(detail_probabilities);

    std::vector<double> v, w;
    distributions.seed(1);
    for (size_t i=0; i<1000000; ++i) {
      double x(distributions.generate(t));
      accumulator_set(x);
      accumulator_p_quantile_1(x);
      accumulator_p_quantile_2(x);
      accumulator_p_quantile_3(x);
      accumulator_p_quantiles_linear(x);
      accumulator_p_quantiles_quadratic(x);
      v.push_back(x);
      quantiles_online.add(x);
      interpolate_quantiles_online.add(x);
    }
    copy(v.begin(), v.end(), std::back_inserter(w));
    sort(w.begin(), w.end());
    double results[] = {
      arithmetic_average(v.begin(), v.end()),
      second_raw_moment(v.begin(), v.end()),
      sample_variance(v.begin(), v.end()),
      sample_variance(v.begin(), v.end()),
      unbiased_variance(v.begin(), v.end()),
      unbiased_variance(v.begin(), v.end()),
      third_raw_moment(v.begin(), v.end()),
      skewness(v.begin(), v.end()),
      fourth_raw_moment(v.begin(), v.end()),
      kurtosis(v.begin(), v.end()),
      *min_element(v.begin(), v.end()),
      *max_element(v.begin(), v.end()),
      sqrt(sample_variance(v.begin(), v.end())/(std::distance(v.begin(), v.end()) - 1)),
      median_in_sorted_container(w.begin(), w.end()),
      median_in_sorted_container(w.begin(), w.end()),
      median_in_sorted_container(w.begin(), w.end()),
      quantile_in_sorted_container(w.begin(), w.end(), 0.25),
      quantile_in_sorted_container(w.begin(), w.end(), 0.5),
      quantile_in_sorted_container(w.begin(), w.end(), 0.75),
#if 0 /* check without heuristic algorithm */
      quantile_in_sorted_container(w.begin(), w.end(), 0.25),
      quantile_in_sorted_container(w.begin(), w.end(), 0.5),
      quantile_in_sorted_container(w.begin(), w.end(), 0.75),
#else
      quantiles_online.get_quantile(.25),
      quantiles_online.get_quantile(.5),
      quantiles_online.get_quantile(.75),
#endif
#if 0 /* check without heuristic algorithm */
      quantile_in_sorted_container(w.begin(), w.end(), 0.25),
      quantile_in_sorted_container(w.begin(), w.end(), 0.5),
      quantile_in_sorted_container(w.begin(), w.end(), 0.75),
#else
      interpolate_quantiles_online.get_quantile(.25),
      interpolate_quantiles_online.get_quantile(.5),
      interpolate_quantiles_online.get_quantile(.75),
#endif
#if 0 /* check without heuristic algorithm */
      quantile_in_sorted_container(w.begin(), w.end(), 0.25),
      quantile_in_sorted_container(w.begin(), w.end(), 0.5),
      quantile_in_sorted_container(w.begin(), w.end(), 0.75),
#else
      interpolate_quantiles_online.get_quantile(.25, true),
      interpolate_quantiles_online.get_quantile(.5, true),
      interpolate_quantiles_online.get_quantile(.75, true),
#endif
    };
    double boost_results[] = {
      boost::accumulators::mean(accumulator_set),
      boost::accumulators::moment<2>(accumulator_set),
      boost::accumulators::variance(accumulator_set),
      boost::accumulators::lazy_variance(accumulator_set),
      boost::accumulators::variance(accumulator_set)*((double)v.size()/(v.size() - 1)),
      boost::accumulators::lazy_variance(accumulator_set)*((double)v.size()/(v.size() - 1)),
      boost::accumulators::moment<3>(accumulator_set),
      boost::accumulators::skewness(accumulator_set),
      boost::accumulators::moment<4>(accumulator_set),
      boost::accumulators::kurtosis(accumulator_set),
      boost::accumulators::min(accumulator_set),
      boost::accumulators::max(accumulator_set),
      boost::accumulators::error_of<boost::accumulators::tag::mean>(accumulator_set),
#if !0
      boost::accumulators::median(accumulator_set),
#else
      boost::accumulators::p_square_quantile_for_median(accumulator_set),
#endif
      boost::accumulators::with_density_median(accumulator_set),
      boost::accumulators::with_p_square_cumulative_distribution_median(accumulator_set),
      boost::accumulators::p_square_quantile(accumulator_p_quantile_1),
      boost::accumulators::p_square_quantile(accumulator_p_quantile_2),
      boost::accumulators::p_square_quantile(accumulator_p_quantile_3),
      boost::accumulators::extended_p_square(accumulator_set)[0],
      boost::accumulators::extended_p_square(accumulator_set)[1],
      boost::accumulators::extended_p_square(accumulator_set)[2],
      boost::accumulators::quantile(accumulator_p_quantiles_linear, boost::accumulators::quantile_probability = .25),
      boost::accumulators::quantile(accumulator_p_quantiles_linear, boost::accumulators::quantile_probability = .5),
      boost::accumulators::quantile(accumulator_p_quantiles_linear, boost::accumulators::quantile_probability = .75),
      boost::accumulators::quantile(accumulator_p_quantiles_quadratic, boost::accumulators::quantile_probability = .25),
      boost::accumulators::quantile(accumulator_p_quantiles_quadratic, boost::accumulators::quantile_probability = .5),
      boost::accumulators::quantile(accumulator_p_quantiles_quadratic, boost::accumulators::quantile_probability = .75),
    };
    for (size_t j=0; j<numberof(boost_results); ++j) {
      double error(sqrt(pow(boost_results[j] - results[j], 2))/(results[j] ? fabs(results[j]) : 1));
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << boost_results[j] << "\t" << results[j] << "\t" << error << std::endl;
      if (!(13 <= j && j <= 18)) assert(error < boost_maximum_permissible_error);
    }
    std::cout << std::endl;
  }
  return 0;
}
