/*
  gsl_running_statistics_distributions.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <cstdlib>
#include <cassert>
#include <gsl/gsl_statistics.h>

#define FUNCTION(a, b) a##_##b
#define BASE double
    
double 
FUNCTION(gsl_stats,third_moment_m) (const BASE data[], 
				    const size_t stride,
				    const size_t n,
				    const double mean)
{
  /* takes a dataset and finds the third_moment */

  long double third_moment = 0;
  size_t i;

  /* we use a recurrence relation to stably update a running value so
     there aren't any large sums that can overflow */

  for (i = 0; i < n; i++)
    {
      const long double x = (data[i * stride] - mean);
      third_moment += (x * x * x - third_moment) / (i + 1);
    }

  return third_moment;
}

double 
FUNCTION(gsl_stats,third_moment) (const BASE data[], const size_t stride, const size_t n)
{
  const double mean = FUNCTION(gsl_stats,mean)(data, stride, n);
  return FUNCTION(gsl_stats,third_moment_m)(data, stride, n, mean);
}
    
double 
FUNCTION(gsl_stats,fourth_moment_m) (const BASE data[], 
				     const size_t stride,
				     const size_t n,
				     const double mean)
{
  /* takes a dataset and finds the fourth_moment */

  long double fourth_moment = 0;
  size_t i;

  /* we use a recurrence relation to stably update a running value so
     there aren't any large sums that can overflow */

  for (i = 0; i < n; i++)
    {
      const long double x = (data[i * stride] - mean);
      fourth_moment += (x * x * x * x - fourth_moment)/(i + 1);
    }

  return fourth_moment;
}

double 
FUNCTION(gsl_stats,fourth_moment) (const BASE data[], const size_t stride, const size_t n)
{
  const double mean = FUNCTION(gsl_stats,mean)(data, stride, n);
  return FUNCTION(gsl_stats,fourth_moment_m)(data, stride, n, mean);
}

#undef FUNCTION
#undef BASE

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_rstat.h>

double
gsl_rstat_third_moment(gsl_rstat_workspace *w)
{
  if (w->n > 0)
    {
      return w->M3 / w->n;
    }
  else
    return 0.0;
} /* gsl_rstat_third_moment() */

double
gsl_rstat_fourth_moment(gsl_rstat_workspace *w)
{
  if (w->n > 0)
    {
      return w->M4 / w->n;
    }
  else
    return 0.0;
} /* gsl_rstat_fourth_moment() */

#include "statistics.hpp"
#include "statistics_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

template <typename InputIterator>
std::vector<std::string> tabs_for_string_list(const InputIterator first, const InputIterator last, const ptrdiff_t offset = 0)
{
  std::vector<std::string> rv;
  InputIterator it(first);
  size_t l = std::string(*it++).size();
  while (it != last)
    l = std::max(l, std::string(*it++).size());
  for (it = first; it != last; ++it) {
    rv.push_back(std::string("\t"));
    for (size_t j=0; j<(l+offset)/8 - (std::string(*it).size()+offset)/8; ++j)
      rv[rv.size()-1] += "\t";
  }
  return rv;
}

double distribution_type_value(const size_t t, gsl_rng *r)
{
  double rv = 0;

  switch (t) {
  case 0:	rv = gsl_rng_uniform(r);	break;
  case 1:	rv = gsl_ran_ugaussian(r);	break;
  case 2:	rv = gsl_ran_gaussian(r, 1.);	break;
  case 3:	rv = gsl_ran_ugaussian_tail(r, 1.5);	break;
  case 4:	rv = gsl_ran_gaussian_tail(r, 1.5, 1.);	break;
  case 5:	rv = gsl_ran_exponential(r, 1.);	break;
  case 6:	rv = gsl_ran_laplace(r, 1.);	break;
  case 7:	rv = gsl_ran_exppow(r, 1., 2.5);	break;
  case 8:	rv = gsl_ran_cauchy(r, 1.);	break;
  case 9:	rv = gsl_ran_rayleigh(r, 1.);	break;
  case 10:	rv = gsl_ran_rayleigh_tail(r, 1., 1.);	break;
  case 11:	rv = gsl_ran_landau(r);	break;
  case 12:	rv = gsl_ran_levy(r, 1., 1.);	break;
  case 13:	rv = gsl_ran_levy_skew(r, 1., 1., 1.);	break;
  case 14:	rv = gsl_ran_gamma(r, 1., 1.);	break;
  case 15:	rv = gsl_ran_flat(r, 0., 1.);	break;
  case 16:	rv = gsl_ran_chisq(r, 1.);	break;
  case 17:	rv = gsl_ran_lognormal(r, 0., 1.);	break;
  case 18:	rv = gsl_ran_fdist(r, 3., 2.);	break;
  case 19:	rv = gsl_ran_tdist(r, 1.);	break;
  case 20:	rv = gsl_ran_beta(r, 2., 2.);	break;
  case 21:	rv = gsl_ran_logistic(r, 1.);	break;
  case 22:	rv = gsl_ran_pareto(r, 1., 1.);	break;
  case 23:	rv = gsl_ran_weibull(r, 1., 1.);	break;
  case 24:	rv = gsl_ran_gumbel1(r, 1., 1.);	break;
  case 25:	rv = gsl_ran_gumbel2(r, 1., 1.);	break;
  case 26:	rv = gsl_ran_poisson(r, 2.5);	break;
  case 27:	rv = gsl_ran_bernoulli(r, .7);	break;
  case 28:	rv = gsl_ran_binomial(r, .5, 9);	break;
  case 29:	rv = gsl_ran_negative_binomial(r, .5, 3.5);	break;
  case 30:	rv = gsl_ran_pascal(r, .5, 3);	break;
  case 31:	rv = gsl_ran_geometric(r, .5);	break;
  case 32:	rv = gsl_ran_hypergeometric(r, 5, 20, 3);	break;
  default:	rv = gsl_ran_logarithmic(r, .7);	break;
  }
  return rv;
}
const char *distribution_names[] = {
  "uniform",
  "ugaussian",
  "gaussian",
  "ugaussian_tail",
  "gaussian_tail",
  "exponential",
  "laplace",
  "exppow",
  "cauchy",
  "rayleigh",
  "rayleigh_tail",
  "landau",
  "levy",
  "levy_skew",
  "gamma",
  "flat",
  "chisq",
  "lognormal",
  "fdist",
  "tdist",
  "beta",
  "logistic",
  "pareto",
  "weibull",
  "gumbel1",
  "gumbel2",
  "poisson",
  "bernoulli",
  "binomial",
  "negative_binomial",
  "pascal",
  "geometric",
  "hypergeometric",
  "logarithmic",
};

//#define USE_RELATIVE_ERROR

int main()
{
#if defined(USE_RELATIVE_ERROR)
  const double maximum_permissible_error_offline(
#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
						 std::numeric_limits<double>::min()
#else
						 //std::numeric_limits<double>::min()
						 //1e-14 // mean
						 1e-9 // skewness
#endif
						 );
  const double gsl_maximum_permissible_error(
					     //std::numeric_limits<double>::min()
					     //1e-14 // mean
					     //1e-13 // variance
					     //1e-12 // skewness
					     1e-10 // kurtosis
					     );
  const double maximum_permissible_error(
#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
                                         //std::numeric_limits<double>::min()
					 1e-12 // skewness
#else
                                         //std::numeric_limits<double>::min()
					 //1e-14 // mean
					 //1e-13 // variance
					 1e-9 // skewness
#endif
                                         );
  const double moving_maximum_permissible_error(
#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
						//std::numeric_limits<double>::min()
						//1e-15 // moving sample variance
						//1e-13 // moving harmonic average
						//1e-12 // moving arithmetic average
						//1e-11 // moving harmonic average
						//1e-7 // moving skewness
						1e-6 // moving kurtosis
#else
						//std::numeric_limits<double>::min()
						//1e-15 // moving geometric average
						//1e-13 // moving unbiased variance
						//1e-10 // moving harmonic average
						//1e-9 // moving unbiased variance
						//1e-8 // moving arithmetic average
						//1e-4 // moving skewness
						1e-3 // moving kurtosis
#endif
						);
#else
  const double maximum_permissible_error_offline(
#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
						 std::numeric_limits<double>::min()
#else
						 //1e-15 // mean
						 //1e-13 // skewness
						 //1e-12 // kurtosis
						 //1e-7 // variance
						 1e-6 // variance
#endif
						 );
  const double gsl_maximum_permissible_error(
					     //std::numeric_limits<double>::min()
					     //1e-13 // mean
					     //1e-12 // kurtosis
					     //1e-8 // variance
					     //1e-7 // variance
					     1e-6 // variance
					     );
  const double maximum_permissible_error(
#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
                                         //std::numeric_limits<double>::min()
					 //1e-15 // skewness
					 //1e-14 // kurtosis
					 //1e-13 // skewness
					 1e-9 // kurtosis
#else
                                         //std::numeric_limits<double>::min()
					 //1e-14 // mean
					 //1e-11 // kurtosis
					 //1e-8 // variance
					 //1e-7 // biased variance
					 1e-6 // variance
#endif
                                         );
  const double moving_maximum_permissible_error(
#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
						//std::numeric_limits<double>::min()
						//1e-16 // moving sample variance
						//1e-15 // moving sample standard deviation
						//1e-14 // moving skewness
						//1e-13 // moving kurtosis
						//1e-5 // moving harmonic average
						1e-4 // moving kurtosis
#else
						//std::numeric_limits<double>::min()
						//1e-15 // moving arithmetic average
						//1e-14 // moving skewness
						//1e-12 // moving kurtosis
						//1e-2 // moving harmonic average
						1e-1 // moving kurtosis
#endif
						);
#endif
  const char *result_names[] = {
    "mean",
    "variance",
    "standard deviation",
    "biased variance",
    "biased standard deviation",
    "skewness",
#if defined(USE_RELATIVE_ERROR)
    "third moment",
#endif
    "kurtosis",
#if defined(USE_RELATIVE_ERROR)
    "fourth moment",
#endif
    "maximum value",
    "minimum value",
    "geometric mean",
    "harmonic mean",
    "root mean square",
  };
  const size_t n(1000000);
  double *a = (double *)malloc(sizeof(double)*n);
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r;

  std::vector<std::string> tabs_for_result_names(tabs_for_string_list(result_names, result_names + numberof(result_names), 1));

  gsl_rng_env_setup();
  r = gsl_rng_alloc(T);

  gsl_rstat_workspace *rstat_p = gsl_rstat_alloc();
  statistics_online<double> statistics_online;

  const char *moving_result_names[] = {
    "moving arithmetic average",
    "moving geometric average",
    "moving harmonic average",
    "moving root mean square",
    "moving sample variance",
    "moving sample standard deviation",
    "moving unbiased variance",
    "moving unbiased standard deviation",
    "moving skewness",
#if defined(USE_RELATIVE_ERROR)
    "moving third moment",
#endif
    "moving kurtosis",
#if defined(USE_RELATIVE_ERROR)
    "moving fourth moment",
#endif
    "moving minimum value",
    "moving maximum value",
  };
  std::vector<double> m(/*10000*/1000);
  moving_statistics_online<double> moving_statistics_online(m.size());

  std::vector<std::string> tabs_for_moving_result_names(tabs_for_string_list(moving_result_names, moving_result_names + numberof(moving_result_names), 1));

  for (size_t t=0; t<33; ++t) {
    std::cout << distribution_names[t] << ":" << std::endl;
    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<n; ++i) {
      a[i] = distribution_type_value(t, r);
      gsl_rstat_add(a[i], rstat_p);
      statistics_online.add(a[i]);
    }

    double gsl_results[] = {
      gsl_stats_mean(a, 1, n),
      gsl_stats_variance(a, 1, n),
      gsl_stats_sd(a, 1, n),
      gsl_stats_variance_with_fixed_mean(a, 1, n, gsl_results[0]),
      gsl_stats_sd_with_fixed_mean(a, 1, n, gsl_results[0]),
      gsl_stats_skew(a, 1, n),
#if defined(USE_RELATIVE_ERROR)
      gsl_stats_third_moment(a, 1, n)/*gsl_stats_skew(a, 1, n)*gsl_results[2]*gsl_results[2]*gsl_results[2]*/,
#endif
      gsl_stats_kurtosis(a, 1, n),
#if defined(USE_RELATIVE_ERROR)
      gsl_stats_fourth_moment(a, 1, n)/*(gsl_stats_kurtosis(a, 1, n)+3)*gsl_results[2]*gsl_results[2]*gsl_results[2]*gsl_results[2]*/,
#endif
      gsl_stats_max(a, 1, n),
      gsl_stats_min(a, 1, n),
    };

    double gsl_running_results[] = {
      gsl_rstat_mean(rstat_p),
      gsl_rstat_variance(rstat_p),
      gsl_rstat_sd(rstat_p),
      gsl_rstat_variance(rstat_p)*(n-1)/n,
      sqrt(pow(gsl_rstat_sd(rstat_p), 2)*(n-1)/n),
      gsl_rstat_skew(rstat_p),
#if defined(USE_RELATIVE_ERROR)
      gsl_rstat_third_moment(rstat_p)/*rstat_p->M3/n*//*gsl_rstat_skew(rstat_p)*gsl_running_results[2]*gsl_running_results[2]*gsl_running_results[2]*/,
#endif
      gsl_rstat_kurtosis(rstat_p),
#if defined(USE_RELATIVE_ERROR)
      gsl_rstat_fourth_moment(rstat_p)/*rstat_p->M4/n*//*(gsl_rstat_kurtosis(rstat_p)+3)*gsl_running_results[2]*gsl_running_results[2]*gsl_running_results[2]*gsl_running_results[2]*/,
#endif
      gsl_rstat_max(rstat_p),
      gsl_rstat_min(rstat_p),
    };

    std::vector<double> v;
    for (size_t i=0; i<n; ++i)
      v.push_back(a[i]);
    double results[] = {
#if !0
      arithmetic_average(v.begin(), v.end()),
#if 0
      unbiased_variance(v.begin(), v.end(), results[0]),
      unbiased_standard_deviation(v.begin(), v.end(), results[0]),
      sample_variance(v.begin(), v.end(), results[0]),
      sample_standard_deviation(v.begin(), v.end(), results[0]),
      skewness(v.begin(), v.end(), results[0], results[2]),
#if defined(USE_RELATIVE_ERROR)
      third_moment(v.begin(), v.end(), results[0])/*skewness(v.begin(), v.end(), results[0], results[2])*results[2]*results[2]*results[2]*/,
#endif
      kurtosis(v.begin(), v.end(), results[0], results[2]),
#if defined(USE_RELATIVE_ERROR)
      fourth_moment(v.begin(), v.end(), results[0])/*(kurtosis(v.begin(), v.end(), results[0], results[2])+3)*results[2]*results[2]*results[2]*results[2]*/,
#endif
      *max_element(v.begin(), v.end()),
      *min_element(v.begin(), v.end()),
#else
      unbiased_variance(v.begin(), v.end()),
      unbiased_standard_deviation(v.begin(), v.end()),
      sample_variance(v.begin(), v.end()),
      sample_standard_deviation(v.begin(), v.end()),
      skewness(v.begin(), v.end()),
#if defined(USE_RELATIVE_ERROR)
      third_moment(v.begin(), v.end())/*skewness(v.begin(), v.end())*results[2]*results[2]*results[2]*/,
#endif
      kurtosis(v.begin(), v.end()),
#if defined(USE_RELATIVE_ERROR)
      fourth_moment(v.begin(), v.end())/*(kurtosis(v.begin(), v.end())+3)*results[2]*results[2]*results[2]*results[2]*/,
#endif
      *std::max_element(v.begin(), v.end()),
      *std::min_element(v.begin(), v.end()),
#endif
      geometric_average(v.begin(), v.end()),
      harmonic_average(v.begin(), v.end()),
      generalized_average(v.begin(), v.end()),
#else
      arithmetic_average(&a[0], &a[n]/*a, a + n*/),
#if 0
      unbiased_variance(&a[0], &a[n]/*a, a + n*/, results[0]),
      unbiased_standard_deviation(&a[0], &a[n]/*a, a + n*/, results[0]),
      sample_variance(&a[0], &a[n]/*a, a + n*/, results[0]),
      sample_standard_deviation(&a[0], &a[n]/*a, a + n*/, results[0]),
      skewness(&a[0], &a[n]/*a, a + n*/, results[0], results[2]),
#if defined(USE_RELATIVE_ERROR)
      third_moment(&a[0], &a[n]/*a, a + n*/, results[0])/*skewness(&a[0], &a[n]/#a, a + n#/, results[0], results[2])*results[2]*results[2]*results[2]*/,
#endif
      kurtosis(&a[0], &a[n]/*a, a + n*/, results[0], results[2]),
#if defined(USE_RELATIVE_ERROR)
      fourth_moment(&a[0], &a[n]/*a, a + n*/, results[0])/*(kurtosis(&a[0], &a[n]/#a, a + n#/, results[0], results[2])+3)*results[2]*results[2]*results[2]*results[2]*/,
#endif
      *std::max_element(&a[0], &a[n]/*a, a + n*/),
      *std::min_element(&a[0], &a[n]/*a, a + n*/),
#else
      unbiased_variance(&a[0], &a[n]/*a, a + n*/),
      unbiased_standard_deviation(&a[0], &a[n]/*a, a + n*/),
      sample_variance(&a[0], &a[n]/*a, a + n*/),
      sample_standard_deviation(&a[0], &a[n]/*a, a + n*/),
      skewness(&a[0], &a[n]/*a, a + n*/),
#if defined(USE_RELATIVE_ERROR)
      third_moment(&a[0], &a[n]/*a, a + n*/)/*skewness(&a[0], &a[n]/#a, a + n#/)*results[2]*results[2]*results[2]*/,
#endif
      kurtosis(&a[0], &a[n]/*a, a + n*/),
#if defined(USE_RELATIVE_ERROR)
      fourth_moment(&a[0], &a[n]/*a, a + n*/)/*(kurtosis(&a[0], &a[n]/#a, a + n#/)+3)*results[2]*results[2]*results[2]*results[2]*/,
#endif
      *std::max_element(&a[0], &a[n]/*a, a + n*/),
      *std::min_element(&a[0], &a[n]/*a, a + n*/),
#endif
      geometric_average(&a[0], &a[n]/*a, a + n*/),
      harmonic_average(&a[0], &a[n]/*a, a + n*/),
      generalized_average(&a[0], &a[n]/*a, a + n*/),
#endif
    };
    std::cout << "gsl_results[i] - results[i]:" << std::endl;
    for (size_t j=0; j<numberof(gsl_results); ++j) {
      double error(sqrt(pow(gsl_results[j] - results[j], 2)));
#if defined(USE_RELATIVE_ERROR)
      if (gsl_results[j] != 0)
	error /= fabs(gsl_results[j]);
#endif
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_results[j] << "\t" << results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error_offline);
    }
    std::cout << std::endl;

    double online_results[] = {
      statistics_online.get_mean(),
      statistics_online.get_unbiased_variance(),
      statistics_online.get_unbiased_standard_deviation(),
      statistics_online.get_sample_variance(),
      statistics_online.get_sample_standard_deviation(),
      statistics_online.get_skewness(),
#if defined(USE_RELATIVE_ERROR)
      statistics_online.get_third_moment()/*statistics_online.get_skewness()*online_results[2]*online_results[2]*online_results[2]*/,
#endif
      statistics_online.get_kurtosis(),
#if defined(USE_RELATIVE_ERROR)
      statistics_online.get_fourth_moment()/*(statistics_online.get_kurtosis()+3)*online_results[2]*online_results[2]*online_results[2]*online_results[2]*/,
#endif
      statistics_online.get_maximum_value(),
      statistics_online.get_minimum_value(),
      statistics_online.get_geometric_mean(),
      statistics_online.get_harmonic_mean(),
      statistics_online.get_root_mean_square(),
    };

    std::cout << "gsl_running_results[i] - gsl_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(gsl_running_results); ++j) {
      double error(sqrt(pow(gsl_running_results[j] - gsl_results[j], 2)));
#if defined(USE_RELATIVE_ERROR)
      if (gsl_results[j] != 0)
	error /= fabs(gsl_results[j]);
#endif
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_running_results[j] << "\t" << gsl_results[j] << "\t" << error << std::endl;
      assert(error < gsl_maximum_permissible_error);
    }
    std::cout << std::endl;

    std::cout << "gsl_results[i] - online_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(gsl_results); ++j) {
      double error(sqrt(pow(gsl_results[j] - online_results[j], 2)));
#if defined(USE_RELATIVE_ERROR)
      if (gsl_results[j] != 0)
	error /= fabs(gsl_results[j]);
#endif
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << gsl_results[j] << "\t" << online_results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
    std::cout << std::endl;

    std::cout << "results[i] - online_results[i]:" << std::endl;
    for (size_t j=0; j<numberof(results); ++j) {
      double error(sqrt(pow(results[j] - online_results[j], 2)));
#if defined(USE_RELATIVE_ERROR)
      if (results[j] != 0)
	error /= fabs(results[j]);
#endif
      std::cout << result_names[j] << ":" << tabs_for_result_names[j] << results[j] << "\t" << online_results[j] << "\t" << error << std::endl;
      if (!(std::isnan(results[j]) && std::isnan(online_results[j]))) assert(error < maximum_permissible_error);
    }
    std::cout << std::endl;

    gsl_rstat_reset(rstat_p);
    statistics_online.reset();

    gsl_rng_set(r, gsl_rng_default_seed);
    for (size_t i=0; i<m.size()*10; ++i) {
      double x(distribution_type_value(t, r));
      m[i % m.size()] = x;
      moving_statistics_online.add(x);
      if (i < m.size()) continue;
      double moving_results[] = {
	arithmetic_average(m.begin(), m.end()),
	geometric_average(m.begin(), m.end()),
	harmonic_average(m.begin(), m.end()),
	generalized_average(m.begin(), m.end(), 2),
	sample_variance(m.begin(), m.end()),
	sample_standard_deviation(m.begin(), m.end()),
	unbiased_variance(m.begin(), m.end()),
	unbiased_standard_deviation(m.begin(), m.end()),
	skewness(m.begin(), m.end()),
#if defined(USE_RELATIVE_ERROR)
	third_moment(m.begin(), m.end())/*skewness(m.begin(), m.end())*moving_results[5]*moving_results[5]*moving_results[5]*/,
#endif
	kurtosis(m.begin(), m.end()),
#if defined(USE_RELATIVE_ERROR)
	fourth_moment(m.begin(), m.end())/*(kurtosis(m.begin(), m.end())+3)*moving_results[5]*moving_results[5]*moving_results[5]*moving_results[5]*/,
#endif
	*min_element(m.begin(), m.end()),
	*max_element(m.begin(), m.end()),
      };
      double moving_online_results[] = {
	moving_statistics_online.get_mean(),
	moving_statistics_online.get_geometric_mean(),
	moving_statistics_online.get_harmonic_mean(),
	moving_statistics_online.get_root_mean_square(),
	moving_statistics_online.get_sample_variance(),
	moving_statistics_online.get_sample_standard_deviation(),
	moving_statistics_online.get_unbiased_variance(),
	moving_statistics_online.get_unbiased_standard_deviation(),
	moving_statistics_online.get_skewness(),
#if defined(USE_RELATIVE_ERROR)
	moving_statistics_online.get_third_moment()/*moving_statistics_online.get_skewness()*moving_online_results[5]*moving_online_results[5]*moving_online_results[5]*/,
#endif
	moving_statistics_online.get_kurtosis(),
#if defined(USE_RELATIVE_ERROR)
	moving_statistics_online.get_fourth_moment()/*(moving_statistics_online.get_kurtosis()+3)*moving_online_results[5]*moving_online_results[5]*moving_online_results[5]*moving_online_results[5]*/,
#endif
	moving_statistics_online.get_minimum_value(),
	moving_statistics_online.get_maximum_value(),
      };

      if (i+1 == m.size()*10) std::cout << "moving_results[i] - moving_online_results[i]:" << std::endl;
      for (size_t j=0; j<numberof(moving_results); ++j) {
	double error(sqrt(pow(moving_results[j] - moving_online_results[j], 2)));
#if defined(USE_RELATIVE_ERROR)
	if (moving_results[j] != 0)
	  error /= fabs(moving_results[j]);
#endif
	if (i+1 == m.size()*10) {
	  std::cout << moving_result_names[j] << ":" << tabs_for_moving_result_names[j] << moving_results[j] << "\t" << moving_online_results[j] << "\t" << error << std::endl;
	  if (j+1 == numberof(moving_results)) std::cout << std::endl;
	}
        if (!(std::isnan(moving_results[j]) || std::isnan(moving_online_results[j]))) assert(error < moving_maximum_permissible_error);
      }
    }
    moving_statistics_online.reset();
  }

  gsl_rng_free(r);
  return 0;
}
