/*
  quantiles.cc

  Written by Taiji Yamada <taiji@aihara.co.jp>
*/
#include <iostream>
#include <vector>
#include <list>
#include <functional>
#if __cplusplus < 201103L
#include <boost/function.hpp>
#endif
#include <cstdlib>      // random(), etc
#include <cassert>
#if __cplusplus < 201103L
#include <boost/utility.hpp>
#endif
#include "quantiles.hpp"
#include "quantiles_online.hpp"

#define numberof(a) (sizeof(a)/sizeof(a[0]))

int main()
{
  const double maximum_permissible_error(1e-12);
  const char *result_names[] = {
    "p=0.25 quantile",
    "p=0.5 quantile",
    "p=0.75 quantile",
    "0-quantile (least)",
    "2-quantile (median)",
    "1-quantile (greatest)",
  };
  std::vector<double> v(1000);
#if 0 /* check without heuristic algorithm */
  moving_quantiles_online<double> quantiles_online(v.size());
#else
  double probabilities[] = { .25, .5, .75, };
  heuristic_quantiles_online<double> quantiles_online(probabilities);
#endif
  srandomdev();
  for (size_t i=0; i<v.size(); ++i) {
    double x((double)random()/((1L<<31)-1));
    v[i] = x;
    quantiles_online.add(x);
  }
  sort(v.begin(), v.end());
  double results[] = {
#if !0
    quantile_in_sorted_container(v.begin(), v.end(), 0.25),
    quantile_in_sorted_container(v.begin(), v.end(), 0.5),
    quantile_in_sorted_container(v.begin(), v.end(), 0.75),
    //*v.begin(),
    least_in_sorted_container(v.begin(), v.end()),
    median_in_sorted_container(v.begin(), v.end()),
    //*(v.begin() + (std::distance(v.begin(), v.end()) - 1)),
    greatest_in_sorted_container(v.begin(), v.end()),
#else
    quantile_in_sorted_container(v.data(), v.data() + v.size(), 0.25),
    quantile_in_sorted_container(v.data(), v.data() + v.size(), 0.5),
    quantile_in_sorted_container(v.data(), v.data() + v.size(), 0.75),
    //v.data()[0],
    least_in_sorted_container(v.data(), v.data() + v.size()),
    median_in_sorted_container(v.data(), v.data() + v.size()),
    //v.data()[(std::distance(v.begin(), v.end()) - 1)],
    greatest_in_sorted_container(v.data(), v.data() + v.size()),
#endif
  };
  double online_results[] = {
    quantiles_online.get_quantile(0.25),
    quantiles_online.get_quantile(0.5),
    quantiles_online.get_quantile(0.75),
    quantiles_online.get_minimum_value(),
    quantiles_online.get_median(),
    quantiles_online.get_maximum_value(),
  };
  for (size_t j=0; j<numberof(online_results); ++j) {
    double error(sqrt(pow(online_results[j] - results[j], 2)));
    std::cout << result_names[j] << ":\t" << online_results[j] << "\t" << results[j] << "\t" << error << std::endl;
#if 0 /* check without heuristic algorithm */
    assert(error < maximum_permissible_error);
#else
    if (!(j<=2 || j==4)) assert(error < maximum_permissible_error);
#endif
  }

  std::vector<double> m(v.size()), s(m.size());
  moving_quantiles_online<double> moving_quantiles_online(v.size());
  for (size_t i=0; i<v.size()*10; ++i) {
    double x((double)random()/((1L<<31)-1));
    m[i % v.size()] = x;
    moving_quantiles_online.add(x);
    if (i < v.size()) continue;
    copy(m.begin(), m.end(), s.begin());
    sort(s.begin(), s.end());
    double moving_results[] = {
#if !0
      quantile_in_sorted_container(s.begin(), s.end(), 0.25),
      quantile_in_sorted_container(s.begin(), s.end(), 0.5),
      quantile_in_sorted_container(s.begin(), s.end(), 0.75),
      //*s.begin(),
      least_in_sorted_container(s.begin(), s.end()),
      median_in_sorted_container(s.begin(), s.end()),
      //*(s.begin() + (std::distance(s.begin(), s.end()) - 1)),
      greatest_in_sorted_container(s.begin(), s.end()),
#else
      quantile_in_sorted_container(s.data(), s.data() + s.size(), 0.25),
      quantile_in_sorted_container(s.data(), s.data() + s.size(), 0.5),
      quantile_in_sorted_container(s.data(), s.data() + s.size(), 0.75),
      //s.data()[0],
      least_in_sorted_container(s.data(), s.data() + s.size()),
      median_in_sorted_container(s.data(), s.data() + s.size()),
      //s.data()[(std::distance(s.begin(), s.end()) - 1)],
      greatest_in_sorted_container(s.data(), s.data() + s.size()),
#endif
    };
    double online_moving_results[] = {
      moving_quantiles_online.get_quantile(0.25),
      moving_quantiles_online.get_quantile(0.5),
      moving_quantiles_online.get_quantile(0.75),
      moving_quantiles_online.get_minimum_value(),
      moving_quantiles_online.get_median(),
      moving_quantiles_online.get_maximum_value(),
    };
    for (size_t j=0; j<numberof(online_moving_results); ++j) {
      double error(sqrt(pow(online_moving_results[j] - moving_results[j], 2)));
      if (i+1 == v.size()*10)
        std::cout << "moving " << result_names[j] << ":\t" << online_moving_results[j] << "\t" << moving_results[j] << "\t" << error << std::endl;
      assert(error < maximum_permissible_error);
    }
  }
  return 0;
}
