/*
  quantiles_online.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _QUANTILES_ONLINE_HPP_
#define _QUANTILES_ONLINE_HPP_
#include "statistics_online.hpp"
#include "quantiles.hpp"
#include <boost/range/begin.hpp>
#include <boost/range/end.hpp>

template <typename T>
class heuristic_quantiles_online : public statistics_online<T> {
public:
  typedef T value_type;
  heuristic_quantiles_online(const T p = .5) : statistics_online<T>(),
    probabilities(1),
    heights(2*probabilities.size() + 3),
    actual_positions(heights.size()),
    desired_positions(heights.size()),
    position_increments(heights.size())
  {
    probabilities[0] = p;
    reset();
  }
  template <typename A>
  heuristic_quantiles_online(const A &a) : statistics_online<T>(),
    probabilities(boost::begin(a), boost::end(a)),
    heights(2*probabilities.size() + 3),
    actual_positions(heights.size()),
    desired_positions(heights.size()),
    position_increments(heights.size())
  {
    reset();
  }
  void reset()
  {
    statistics_online<T>::reset();
    size_t n_quantiles(probabilities.size()), n_markers(heights.size());
    for (size_t j=0; j<n_markers; ++j)
      actual_positions[j] = j + 1;
    position_increments[0] = 0;
    position_increments[n_markers - 1] = 1;
    for (size_t i=0; i<n_quantiles; ++i)
      position_increments[2*i + 2] = probabilities[i];
    for (size_t i=0; i<n_quantiles+1; ++i)
      position_increments[2*i + 1] = .5*(position_increments[2*i + 0] + position_increments[2*i + 2]);
    for (size_t j=0; j<n_markers; ++j)
      desired_positions[j] = 1 + 2*(n_quantiles + 1)*position_increments[j];
  }
  void add(const T v)
  {
    unsigned long long count(statistics_online<T>::get_count());
    statistics_online<T>::add(v);
    size_t n_markers(2*probabilities.size() + 3);
    if (count < n_markers) {
      heights[count] = v;
      if (++count == n_markers)
	std::sort(heights.begin(), heights.end());
    }
    else {
      typename std::vector<T>::size_type hid;
      if (v < heights[0])
	heights[(hid = 1) - 1] = v;
      else if (!(v < heights[n_markers - 1]))
	heights[(hid = n_markers - 1)] = v;
      else
	hid = std::distance(heights.begin(), std::upper_bound(heights.begin(), heights.end(), v));

      for (size_t j=hid; j<n_markers; ++j)
	++actual_positions[j];
      for (size_t j=0; j<n_markers; ++j)
	desired_positions[j] += position_increments[j];

      for (size_t j=1; j<n_markers-1; ++j) {
	value_type da[] = {
	  actual_positions[j-1] - actual_positions[j],
	  actual_positions[j+1] - actual_positions[j],
	}, ha[] = {
	  (heights[j-1] - heights[j])/da[0],
	  (heights[j+1] - heights[j])/da[1],
	}, d = desired_positions[j] - actual_positions[j];

	if ((!(-1 < d) && da[0] < -1) || (!(d < 1) && 1 < da[1])) {
	  int sign_d = static_cast<int>(d/std::abs(d));
	  value_type h = heights[j] + sign_d/(da[1] - da[0])*((da[1] - sign_d)*ha[0] + (sign_d - da[0])*ha[1]);

	  if (heights[j - 1] < h && h < heights[j + 1])
	    heights[j] = h;
	  else {
	    if (d < 0) heights[j] -= ha[0];
	    if (d > 0) heights[j] += ha[1];
	  }
	  actual_positions[j] += sign_d;
	}
      }
    }
    results.clear();
    for (size_t i=1; i<probabilities.size()+1; ++i)
      results.push_back(heights[i*2]);
  }
  T get_quantile(const T probability, const bool option_quadratic = false) const
  {
    T rv;
    typename std::vector<value_type>::const_iterator pit = std::lower_bound(probabilities.begin(), probabilities.end(), probability);
    typename std::vector<value_type>::size_type pid = std::distance(probabilities.begin(), pit);
    if (probability < *probabilities.begin() || *(probabilities.end() - 1) < probability)
      rv = std::numeric_limits<T>::quiet_NaN();
    else if (*pit == probability)
      rv = results[pid];
    else {
      typename std::vector<value_type>::const_iterator hit = results.begin() + pid;
      if (!option_quadratic) {
	value_type p[] = { *(pit - 1), *pit, }, h[] = { *(hit - 1), *hit, };
	value_type c[] = { (h[1] - h[0])/(p[1] - p[0]), h[1] - c[0]*p[1], };
	rv = c[0]*probability + c[1];
      }
      else {
	if (!((pid == 1 || !(probability - *(pit - 1) < *pit - probability)) &&
	      pid != probabilities.size() - 1)) {
	  --pit;
	  --hit;
	}
	value_type p[] = { *(pit - 1), *pit, *(pit + 1), }, h[] = { *(hit - 1), *hit, *(hit + 1), };
	value_type hp[] = { (h[1] - h[0])/(p[1] - p[0]), (h[2] - h[1])/(p[2] - p[1]), };
	value_type pp[] = { (p[1]*p[1] - p[0]*p[0])/(p[1] - p[0]), (p[2]*p[2] - p[1]*p[1])/(p[2] - p[1]), };
	value_type c[] = { (hp[1] - hp[0])/(pp[1] - pp[0]), hp[1] - c[0]*pp[1], h[2] - c[0]*p[2]*p[2] - c[1]*p[2], };
	rv = c[0]*probability*probability + c[1]*probability + c[2];
      }
    }
    return rv;
  }
  T get_median() const
  {
    return get_quantile(.5, false);
  }
private:
  std::vector<value_type> probabilities, heights, actual_positions, desired_positions, position_increments, results;
  /*
    References:

    [1] K. E. E. Raatikainen, ``Simultaneous estimation of several
    percentiles,'' Simulation, Vol. 49, No. 4, pp. 159--164, 1987.

    [2] Raj Jain and Imrich Chlamtac, ``The P^2 Algorithm for Dynamic
    Calculation of Quantiles and Histograms Without Storing
    Observations,'' Communications of the ACM, Vol. 28, No. 10, pp. 1076--1085, October 1985.

    [3] ftp://ftp.gnu.org/gnu/gsl/, 2016.

    [4] http://boost-sandbox.sourceforge.net/libs/accumulators/doc/html/accumulators/user_s_guide/the_statistical_accumulators_library.html, 2016.
  */
};

template <typename T>
class moving_quantiles_online : public moving_statistics_online<T> {
public:
  moving_quantiles_online(const size_t size = 100) : moving_statistics_online<T>(size), sorted(false)
  {
    sorted_storage = new T[size];
  }
  virtual ~moving_quantiles_online()
  {
    delete [] sorted_storage;
  }
  void add(const T v)
  {
    moving_statistics_online<T>::add(v);
    sorted = false;
  }
  T get_median()
  {
    const size_t size(moving_statistics_online<T>::size);
    const size_t index(moving_statistics_online<T>::get_index());
    const bool initial(moving_statistics_online<T>::is_initial());
    if (!sorted)
      sort();
    return median_in_sorted_container(sorted_storage, sorted_storage + (initial ? index : size));
  }
  T get_quantile(const T p)
  {
    const size_t size(moving_statistics_online<T>::size);
    const size_t index(moving_statistics_online<T>::get_index());
    const bool initial(moving_statistics_online<T>::is_initial());
    if (!sorted)
      sort();
    return quantile_in_sorted_container(sorted_storage, sorted_storage + (initial ? index : size), p);
  }
  void reset() { moving_statistics_online<T>::reset(); sorted = false; }
private:
  void sort()
  {
    const size_t size(moving_statistics_online<T>::size);
    const size_t index(moving_statistics_online<T>::get_index());
    const bool initial(moving_statistics_online<T>::is_initial());
    std::copy(moving_arithmetic_average_online<T>::get_storage(), moving_arithmetic_average_online<T>::get_storage() + (initial ? index : size), sorted_storage);
    std::sort(sorted_storage, sorted_storage + (initial ? index : size));
    sorted = true;
  }
  bool sorted;
  T *sorted_storage;
};

#endif
